#!/usr/bin/env ruby
# $Id: rqgrep.rb,v 1.6 2003/07/24 13:17:43 noir Exp $
# Copyright (C) 2003 Hiroyuki KUROSAKI <noir@st.rim.or.jp>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
$KCODE = 'E'
require 'kconv'
require "qte"
require "qpe"
include Qte
include Qpe

module RQGrep
  VERSION = "0.0.5"
  COPYRIGHT = "Copyright (C) 2003 Hiroyuki KUROSAKI"

  class SearchThread <QObject
    def initialize(q_pattern, q_dirglob, parent = nil)
      super(parent, "Search Thread")
      @codec = QTextCodec::codecForName("eucJP")
      @pattern = @codec.fromUnicode(q_pattern).to_str
      @dirglob = q_dirglob.utf8().to_str
      @found = RSignal.new("Object")
      @finished = RSignal.new()
      @timer = QTimer.new(self)
      connect(@timer, QSIGNAL("timeout()"), self, "get_each_line_from_files")
    end
    attr_reader :found, :finished

    def pattern=(qstr)
      @pattern = @codec.fromUnicode(qstr).to_str
    end

    def pattern
      @pattern
    end

    def dirglob=(qstr)
      @dirglob = qstr.utf8().to_str
    end

    def dirglob
      @dirglob
    end

    def start
      @pat_regexp = Regexp.new(pattern)
      @files_queue = Dir.glob(dirglob)
      @current_file = nil
      @timer.start(0, false)
    end

    def stop
      @timer.stop
      @finished.send
    end

    def get_each_line_from_files
      if !@current_file then
	if @files_queue.size == 0 then
	  stop
	else
	  begin
	    @current_file = File.open(@files_queue.shift, "r")
	    @lnum = 0
	  rescue
	    @current_file = nil
	  end
	end
      end
      if @current_file then
	if line = @current_file.gets then
	  @lnum += 1
	  line = Kconv::toeuc(line.chomp)
	  if @pat_regexp.match(line) then
	    mdata = MatchedData.new(@current_file.path,
				    @lnum,
				    @codec.toUnicode(line).utf8().to_str)
	    @found.sendWith(mdata)
	  end
	else
	  @current_file.close
	  @current_file = nil
	end
      end
    end

  end

  class MatchedData
    def initialize(data, lnumber = nil, line = nil)
      if !lnumber && !line && /^([^:]+):([0-9]+):(.*)$/u =~ data then
	@filename = $1
	@lnumber = $2.to_i
	@line = $3
      else
	@filename = data
	@lnumber = lnumber
	@line = line
      end
    end
    attr_reader :filename, :lnumber, :line

    def to_s
      filename + ":" + lnumber.to_s + ":" + line
    end

  end

  class ExternalEditor
    def initialize(progname = "textedit")
      @progname = progname
    end
    attr_reader :progname

    def invoke(mdata)
      env = QCopEnvelope.new("QPE/System", "execute(QString,QString)")
      env << QString.new(progname) << QString::fromUtf8(mdata.filename)
      env = nil
      GC.start
    end

  end

  class Results <QObject
    include Enumerable

    def initialize(parent = nil)
      super(parent, "Search Results")
      @codec = QTextCodec::codecForName("eucJP")
      @resary = []
      @view = QListBox.new(parent, "Search Results List Box")
      @editor = ExternalEditor.new
      connect(@view, QSIGNAL("clicked(QListBoxItem*)"), self, "invoke_editor")
      connect(@view, QSIGNAL("returnPressed(QListBoxItem*)"), self, "invoke_editor")
    end

    attr_reader :view

    def invoke_editor(lbitem)
      if lbitem then
	@editor.invoke(MatchedData.new(lbitem.text().utf8().to_str))
      end
    end

    def each(&block)
      @resary.each {|item|
	yield(item)
      }
    end

    def clear
      @resary.clear
      @view.clear()
    end

    def size
      @resary.size
    end

    def push(mdata)
      @resary.push(mdata)
      @view.insertItem(QString::fromUtf8(mdata.to_s))
    end

    def pop
      @resary.pop
    end

    def shift
      @resary.shift
    end
    
    def unshift(qstr)
      @resary.unshift(qstr)
    end

  end

  class Main <QMainWindow
    def initialize
      super(nil, "RQGrep")
      setCaption("RQGrep")
      @config = Config.new(QString.new('rqgrep'))
      setupMenuBar
      setupCentralWidget
      @stext.setFocus()
      @is_searching = false
      statusBar().addWidget(QLabel.new("", statusBar(), "Status Bar Placeholder"), 10)
      statusBar().message("Ready.")
    end

    def setupMenuBar
      @menubar = QPEMenuBar.new(self)
      @filemenu = QPopupMenu.new(self, "File Menu")
      @filemenu.insertSeparator()
      @filemenu.insertItem("&Quit", self, "close()", CTRL+Key_Q)
      @editmenu = QPopupMenu.new(self, "Edit Menu")
      @editmenu.insertItem("&Undo", self, "undo", CTRL+Key_Z)
      @editmenu.insertSeparator()
      @editmenu.insertItem("Cu&t", self, "cut", CTRL+Key_X)
      @editmenu.insertItem("&Copy", self, "copy", CTRL+Key_C)
      @editmenu.insertItem("&Paste", self, "paste", CTRL+Key_V)
      menuBar().insertItem("&File", @filemenu)
#      menuBar().insertItem("&Edit", @editmenu)
      @helpmenu = QPopupMenu.new(self, "Help Menu")
      @helpmenu.insertItem("&About", self, "about")
      menuBar().insertSeparator()
      menuBar().insertItem("&Help", @helpmenu)
    end

    def setupCentralWidget
      @cwidget = QWidget.new(self, "Central Widget")

      @clayout = QGridLayout.new(@cwidget, 3, 3, 0, 2)
      @clayout.setColStretch(1, 0)
      @clayout.setRowStretch(2, 0)

      @slabel = QLabel.new("&Search:", @cwidget)
      @stext = QComboBox.new(true, @cwidget, "Search Text")      
      cb_setup(@stext, 'stext')
      @slabel.setBuddy(@stext)
      # text "&Cancel" ... to adjust layout.
      @sbutton = QPushButton.new("&Cancel", @cwidget, "Go/Cancel Button")
      connect(@sbutton, QSIGNAL("clicked()"), self, "toggle_search")

      @clayout.addItem(QWidgetItem.new(@slabel))
      @clayout.addItem(QWidgetItem.new(@stext))
      @clayout.addItem(QWidgetItem.new(@sbutton))
      @sbutton.setFixedSize(@sbutton.sizeHint())
      @sbutton.setText("&Go")

      @dlabel = QLabel.new("&Dir:", @cwidget)
      @dtext = QComboBox.new(true, @cwidget, "Dir Glob Text")
      @dlabel.setBuddy(@dtext)
      @dtext.setMinimumWidth(10)

      @clayout.addItem(QWidgetItem.new(@dlabel))
      @clayout.addMultiCellWidget(@dtext, 1, 1, 1, 2)

      cb_setup(@dtext, 'dtext')
      cb_set_defaults(@dtext,
		      [ ENV['HOME'] + '/Documents/Text_Files/**/*.txt',
			'/mnt/card/Documents/Text_Files/**/*.txt',
			'/mnt/cf/Documents/Text_Files/**/*.txt' ])

      @results = Results.new(@cwidget)
      @clayout.addMultiCellWidget(@results.view, 2, 2, 0, 2)

      setCentralWidget(@cwidget)
    end

    def undo
    end

    def cut
    end
    
    def copy
    end

    def paste
    end

    def about
      QMessageBox::about(self, "RQGrep " + "Ver. " + RQGrep::VERSION,
			 RQGrep::COPYRIGHT)
    end

#    attr_reader :canceled

    def toggle_search
      @sbutton.setEnabled(false)
      if @is_searching then
	@is_searching = false
	stop_search
	@sbutton.setText("&Go")
	statusBar().message("Stop.")
      else
	if @stext.currentText().length > 0 then
	  @is_searching = true
	  cb_update(@stext)
	  cb_update(@dtext)
	  save_history
	  start_search
	  @sbutton.setText("&Cancel")
	  statusBar().message("Searching...")
	end
      end
      @sbutton.setEnabled(true)
    end

    def finish_search
      @is_searching = false
      @sbutton.setText("&Go")
      statusBar().message("Finished.")
    end

    def start_search
      unless @sthread
	@sthread = SearchThread.new(@stext.currentText(),
				    @dtext.currentText(), self)
	connect(@sthread.found, @results, "push")
	connect(@sthread.finished, self, "finish_search")
      else
	@sthread.pattern = @stext.currentText()
	@sthread.dirglob = @dtext.currentText()
      end
      @results.clear
      @sthread.start
    end

    def stop_search
      if @sthread then
	@sthread.stop
      end
    end

    def cb_setup(cb, entryname)
      @config.setGroup(QString.new(entryname))
      (0 ... 5).each do |i|
	val = @config.readEntry(QString.new(entryname + i.to_s))
	if val.length != 0 then
	  cb.insertItem(val)
	end
      end
    end

    def cb_save(cb, entryname)
      @config.setGroup(QString.new(entryname))
      (0 ... cb.count).each do |i|
	if cb.count == 5 then
	  break
	end
	@config.writeEntry(QString.new(entryname + i.to_s), cb.text(i))
      end
    end

    def cb_update(cb)
      qstr = cb.currentText()
      i = -1
      (0 ... cb.count).each do |j|
	if cb.text(j).compare(qstr) == 0 then
	  i = j
	end
      end
      if i != -1 then
	cb.removeItem(i)
      end
      cb.insertItem(qstr, 0)
      cb.setCurrentItem(0)
    end

    def cb_set_defaults(cb, values)
      already_exists = Hash.new
      (0 ... cb.count).each do |i|
	values.each do |v|
	  if cb.text(i).compare(QString.new(v)) == 0 then
	    already_exists[v] = true
	  end
	end
      end
      values.each do |v|
	unless already_exists[v]
	  cb.insertItem(v)
	end
      end
    end

    def save_history
      cb_save(@stext, 'stext')
      cb_save(@dtext, 'dtext')
    end
  end
end

if $0 == __FILE__ then
  a = QPEApplication.new([$0] + ARGV)
  mw = RQGrep::Main.new
  a.showMainWidget(mw)
  a.exec
end
